<?php

require_once __DIR__ . '/vendor/autoload.php';

use League\CLImate\CLImate;

///////////////////////////////////////////////////////////////

class PagheOpenUpdater {

	const ANAG_OLD_VER = '01.45.00'; // This value will be changed by Robo
	const ANAG_NEW_VER = '01.45.00'; // This value will be changed by Robo
	const PAGHE_OLD_VER = '02.40.00'; // This value will be changed by Robo
	const PAGHE_NEW_VER = '02.41.00'; // This value will be changed by Robo
	const DB_RUNTIME_PATH_LINUX = './firebird_data/'; // This value will be changed by Robo
	const DEF_USER = 'sysdba';
	
	const DEF_PASS = 'masterkey';
	const DEF_HOST = 'localhost';
	const DEF_PATH = '/firebird_data';
	
	const FIREBIRD_CONN_LOCAL = 1;
	const FIREBIRD_CONN_EMBEDDED = 2;
	const FIREBIRD_CONN_SERVER = 3;
	const FIREBIRD_CONN_NATIVE = 4;

	private $isql = null;	
	private $climate = null;	
	private $array1 = null;
	private $array2 = null;
	private $firebird_conn_type = null;
	private $current_db_versions = [];
	
	function __construct() {
		$this->array1 = [
			'anag.fdb',
			'common.fdb',
			'blob.fdb'
		];
		
		$this->array2 = [
			'paghe.fdb',
			'presenze.fdb'
		];
	}
	
	public function run(){		
		$this->user = self::DEF_USER;
		$this->pass = self::DEF_PASS;
		
		$anag_old_version = self::ANAG_OLD_VER;
		$app_old_version = self::PAGHE_OLD_VER;
		$anag_new_version = self::ANAG_NEW_VER;
		$app_new_version = self::PAGHE_NEW_VER;
				
		$this->climate = new CLImate();
		
		$input = $this->climate->input('Specificare il tipo di connessione');
		echo self::FIREBIRD_CONN_LOCAL .') local' . PHP_EOL;
		echo self::FIREBIRD_CONN_EMBEDDED .') embedded' . PHP_EOL;
		echo self::FIREBIRD_CONN_SERVER .') server' . PHP_EOL;
		echo self::FIREBIRD_CONN_NATIVE .') native' . PHP_EOL;

		$this->firebird_conn_type = $input->prompt();
		
		// solo se connection type è SERVER
		if ($this->firebird_conn_type == self::FIREBIRD_CONN_SERVER){
			$b = $this->checkFirebird(gethostbyname(self::DEF_HOST));
			if(!$b){
				$this->climate->error('Firebird non è in esecuzione su ' . self::DEF_HOST);
				die('Quit' . PHP_EOL);
			}
		}

		//$fdb_path = self::DEF_PATH;
		$fdb_path = null;
		$b = $this->areFdb($fdb_path);
		if(!$b){
			$continue = false;
			while(!$continue){
				$input = $this->climate->input('Specificare il percorso delle tabelle da aggiornare');
				$input->defaultTo(self::DEF_PATH);
				$fdb_path = $input->prompt(); 
				$fdb_path = $this->removeQuotes($fdb_path);
				$b = $this->areFdb($fdb_path);
				if(!$b){
					$this->climate->error('Il percorso ' . $fdb_path . ' non contiene le tabelle.' . PHP_EOL);
				}else{
					$continue = true;
					// se connection type è EMBEDDED eseguo comandi:
					if ($this->firebird_conn_type == self::FIREBIRD_CONN_EMBEDDED){
						$path = $fdb_path . '/..';
						$cmd1 = 'LD_LIBRARY_PATH=' . $path . '/firebird';
						$cmd2 = 'FIREBIRD=' . $path . '/firebird';
						$cmd3 = 'PATH=$PATH:' . $path . '/firebird/bin';

						putenv($cmd1);
						putenv($cmd2);
						putenv($cmd3);
					}
				}
			}
		}
		
		$this->current_db_versions = $this->getVersions($fdb_path);
 
		// Before the update...
		$this->areFdbVer2($fdb_path, $this->array1, $anag_old_version);
		$this->areFdbVer2($fdb_path, $this->array2, $app_old_version);

		$files = $this->getUpdateFiles(__DIR__);
		if(count($files)==0){
			$this->climate->green('ERROR: impossibile individuare i file di aggiornamento');
			die('QUIT' . PHP_EOL);
		}

		foreach($files as $sql_file){
			$this->runUpdate($sql_file, $fdb_path);
		}
		// ..after the update
		$this->current_db_versions = $this->getVersions($fdb_path);
		$b3 = $this->areFdbVer2($fdb_path, $this->array1, $anag_new_version);
		$b4 = $this->areFdbVer2($fdb_path, $this->array2, $app_new_version);
		if($b3 && $b4){
			$this->climate->green('Aggiornamento eseguito con successo.');
		}else{
			$this->climate->green('Si sono verificati alcuni errori.');
		}
	}

	private function getVersions($fdb_path){
		$vers = array();
		$db_get_version_file = __DIR__ . DIRECTORY_SEPARATOR . "Get_Db_Version.sql";
		$info = new \SplFileInfo($db_get_version_file);		
		$out_file = __DIR__ . '/' . $info->getBasename('.' . $info->getExtension()) . "-isql.log"; // Get_Db_Version-isql.log
		$this->runQueryFromFile($db_get_version_file, $out_file, $fdb_path);
		$content = file_get_contents($out_file);
		
		preg_match_all('/(\d\d\.\d\d.\d\d)/', $content, $matches, PREG_SET_ORDER, 0);
		$vers['anag'] = $matches[0][0];
		$vers['paghe'] = $matches[1][0];
		$vers['presenze'] = $matches[2][0];
		$vers['blob'] = $matches[3][0];
		$vers['common'] = $matches[4][0];
		
		return $vers;
	}
			
	private function runUpdate($sql_file, $fdb_path){
		$info = new \SplFileInfo($sql_file);		
		$out_file = __DIR__ . DIRECTORY_SEPARATOR . $info->getBasename('.' . $info->getExtension()) . "-isql.log";
		$this->runQueryFromFile($sql_file, $out_file, $fdb_path);
	}
	
	private function runQueryFromFile($sql_file_orig, $out_file, $fdb_path){
		$path_parts = pathinfo($sql_file_orig);
		$file_name = $path_parts['basename'];
		
		$sql_file = sys_get_temp_dir() . DIRECTORY_SEPARATOR . $file_name;
		file_put_contents($sql_file, file_get_contents($sql_file_orig));
		
		$this->findAndReplace([$sql_file], $fdb_path);
		
		if (is_file($out_file)){
			file_put_contents($out_file, '');
		}
		
		$isql_exe = $this->enclose($this->getIsqlExe());
		$credentials = "-u " . $this->user . " -p " . $this->pass;
		$cmd = $isql_exe . " " . $credentials . " -ch UTF8 -q -s 3 -m -i " . $this->enclose($sql_file) . " -o " . $this->enclose($out_file);
		$exit_code = $this->runCmd($cmd);	
		$this->climate->debug('Log file is: ' . $out_file);
		if ($exit_code) {
			$this->climate->br();
			$this->climate->error("ERROR: the last command returned the exit code $exit_code");
		}
		$this->climate->info('Procedura di aggiornamento completata con successo' . PHP_EOL);				
	}

	protected function checkFile($file) {
		if (!is_file($file)) {
			$error = "ERROR: file '$file' not found !!";
			throw new \RuntimeException($error);
		}
		return true;
	}

	private function isFdbVer($fullpath, $expected_ver){
	    $b = false;

		$path_parts = pathinfo($fullpath);
		$db_name = $path_parts['filename'];
		
		$actual_ver = $this->current_db_versions[$db_name];
		
		/*
	    $host_name = 'localhost';
		if ($this->firebird_conn_type == self::FIREBIRD_CONN_EMBEDDED){
			$host_name = 'embedded';
		}
		
	    $query = "SELECT VALORE FROM DBINFO WHERE ID = 1";
	    $conn = new FirebirdConn($host_name, $fullpath, $this->firebird_conn_type);
	    $conn->open();
	    $records = $conn->query($query);
	    $row = $records[0];
	    $actual_ver = $row['VALORE'];
	    $conn->close();
		*/
		
	    if($actual_ver == $expected_ver){
	        $b = true;
	    }else{
	        $msg = "fdb: " . $fullpath . " | actual ver: " . $actual_ver . " | expected ver: " . $expected_ver;
	        $this->climate->error('ERROR: ' . $msg);
	    }
		
		
	    return $b;
	}
	
	private function getIsqlExe() {
		if(!$this->isql || !is_file($this->isql)){
			$this->isql  = $this->askExe("isql path:");
		}
		
		$this->isql = $this->removeQuotes($this->isql);
		$this->checkFile($this->isql);
		chmod($this->isql, 0777);
		return $this->isql;
	}
	
	private function runCmd(string $cmd) {
		$this->climate->yellow($cmd); // $this->climate->whisper($cmd);
		exec($cmd, $output, $exit_code);
	 	$out_str = implode(PHP_EOL, $output);	 
		if ($exit_code) {
			$this->climate->br();
			$this->climate->dump($output);
		}
		return $exit_code;
	}
		
	private function checkFirebird($ip = '127.0.0.1', $port = 3050) {
	    $fp = @fsockopen($ip, $port, $errno, $errstr, 0.1);
	    if (!$fp) {
	        return false;
	    } else {
	        fclose($fp);
	        return true;
	    }
	}
		
	private function askExe($msg) {
		$input = $this->climate->input($msg);
		$exe = $input->prompt();
		return $exe;
	}
		
	private function enclose(string $str) {
		return "\"" . $str . "\""; // TODO: valutare se è meglio usare escapeshellarg()
	}
	
	private function areFdbVer2($path, array $array, $ver){
		foreach ($array as $filename) {
			$fullpath = $path . '/' . $filename;
			if (!is_file($fullpath)) {
				die("QUIT: missing file: " . $fullpath . PHP_EOL);
			}else{
				$b = $this->isFdbVer($fullpath, $ver);
				if(!$b){
					return false;
				}
			}
		}
		return true;
	}
	
	private function areFdb($path) {
		$array = array_merge($this->array1, $this->array2);
		foreach ($array as $filename) {
			$fullpath = $path . '/' . $filename;
			if (!is_file($fullpath)) {
				return false;
			}
		}
		return true;
	}
	
	private function getUpdateFiles($fdb_path){
		$files = [];
		try {
			$regex = "/update_db_(.*)\\.sql/i";	
			$files_found = $this->getFileByPattern($fdb_path, $regex);
			foreach ($files_found as $file) {
				$files[] = $file->getPathName();
			}
		
			sort($files);
		} catch(\Exception $e) {
			$this->climate->error('Errore durante la lettura della cartella "'. $fdb_path .'"');
			$this->climate->error($e->getMessage());
		}

		return $files;
	}
	
	private function findAndReplace($files, $fdb_path){
		foreach ($files as $filename) {
			$b = $this->changeDsnPathOnFile($filename, $fdb_path);
		}
	}		

	private function changeDsnPathOnFile($filename, $fdb_path) {		
		$this->checkFile($filename);
		$file = $info = new \SplFileObject ($filename);
		$lines = file($file->getPathname()); // Read file content to array
		// TODO: aggiungere eventuale slash finale a $fdb_path
		// substr("testers", -1); // returns "s"
		
		$host_name = null;
		if ($this->firebird_conn_type == self::FIREBIRD_CONN_LOCAL){
			$host_name = 'localhost';
		} else if ($this->firebird_conn_type == self::FIREBIRD_CONN_SERVER){
			$host_name = 'localhost';
		}
		
		if (substr($fdb_path, -1) !== '/'){
			$fdb_path .= '/';
		}
				
		$n = 0;
		foreach ($lines as $line) {
			$pos = stripos($line, self::DB_RUNTIME_PATH_LINUX);
			if ($pos !== false) {
				$full_path = null;
				if ($host_name !== null){
					$full_path = $host_name . ':';
				}
				
				$full_path = $full_path . $fdb_path;
				$new_line = str_ireplace(self::DEF_HOST . ':' . self::DB_RUNTIME_PATH_LINUX, $full_path, $line);
				$lines[$n] = $new_line;
			}
			$n ++;
		}
		return file_put_contents($file->getPathname(), implode("\n", $lines));
	}
	
	private function removeQuotes($str){
		$str = str_replace('"', '', $str);
		$str = str_replace("'", '', $str);
		
		return $str;
	}
	
	public function getFileByPattern($path='.', $regex=''){ // $regex example '/^.*\.(php|dat)$/' oppure /^.+\.php$/i
		$iterator = new \RecursiveDirectoryIterator($path);
		$filter = new \RegexIterator($iterator->getChildren(), $regex);
		return $filter;
	}
			
} // end class		

////////////////// Main

$obj = new PagheOpenUpdater();
$obj->run();
